@@ -4,6 +4,7 @@ import os
 import platform
 import sys
 import threading
+import zlib
 from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from datetime import datetime
@@ -12,6 +13,7 @@ from getpass import getpass
 from html import escape
 from inspect import isclass
 from itertools import islice
+from math import ceil
 from time import monotonic
 from types import FrameType, ModuleType, TracebackType
 from typing import (
@@ -43,9 +45,10 @@ else:
 
 from . import errors, themes
 from ._emoji_replace import _emoji_replace
+from ._export_format import CONSOLE_HTML_FORMAT, CONSOLE_SVG_FORMAT
 from ._log_render import FormatTimeCallable, LogRender
 from .align import Align, AlignMethod
-from .color import ColorSystem
+from .color import ColorSystem, blend_rgb
 from .control import Control
 from .emoji import EmojiVariant
 from .highlighter import NullHighlighter, ReprHighlighter
@@ -82,159 +85,22 @@ class NoChange:
 
 NO_CHANGE = NoChange()
 
+try:
+    _STDIN_FILENO = sys.__stdin__.fileno()
+except Exception:
+    _STDIN_FILENO = 0
 try:
     _STDOUT_FILENO = sys.__stdout__.fileno()
 except Exception:
     _STDOUT_FILENO = 1
-
 try:
     _STDERR_FILENO = sys.__stderr__.fileno()
 except Exception:
     _STDERR_FILENO = 2
 
-_STD_STREAMS = (_STDOUT_FILENO, _STDERR_FILENO)
+_STD_STREAMS = (_STDIN_FILENO, _STDOUT_FILENO, _STDERR_FILENO)
+_STD_STREAMS_OUTPUT = (_STDOUT_FILENO, _STDERR_FILENO)
 
-CONSOLE_HTML_FORMAT = """\
-<!DOCTYPE html>
-<head>
-<meta charset="UTF-8">
-<style>
-{stylesheet}
-body {{
-    color: {foreground};
-    background-color: {background};
-}}
-</style>
-</head>
-<html>
-<body>
-    <code>
-        <pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace">{code}</pre>
-    </code>
-</body>
-</html>
-"""
-
-CONSOLE_SVG_FORMAT = """\
-<svg width="{total_width}" height="{total_height}" viewBox="0 0 {total_width} {total_height}"
-     xmlns="http://www.w3.org/2000/svg">
-    <style>
-        @font-face {{
-            font-family: "Fira Code";
-            src: local("FiraCode-Regular"),
-                 url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff2/FiraCode-Regular.woff2") format("woff2"),
-                 url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff/FiraCode-Regular.woff") format("woff");
-            font-style: normal;
-            font-weight: 400;
-        }}
-        @font-face {{
-            font-family: "Fira Code";
-            src: local("FiraCode-Bold"),
-                 url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff2/FiraCode-Bold.woff2") format("woff2"),
-                 url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff/FiraCode-Bold.woff") format("woff");
-            font-style: bold;
-            font-weight: 700;
-        }}
-        span {{
-            display: inline-block;
-            white-space: pre;
-            vertical-align: top;
-            font-size: {font_size}px;
-            font-family:'Fira Code','Cascadia Code',Monaco,Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace;
-        }}
-        a {{
-            text-decoration: none;
-            color: inherit;
-        }}
-        .blink {{
-           animation: blinker 1s infinite;
-        }}
-        @keyframes blinker {{
-            from {{ opacity: 1.0; }}
-            50% {{ opacity: 0.3; }}
-            to {{ opacity: 1.0; }}
-        }}
-        #wrapper {{
-            padding: {margin}px;
-            padding-top: 100px;
-        }}
-        #terminal {{
-            position: relative;
-            display: flex;
-            flex-direction: column;
-            align-items: center;
-            background-color: {theme_background_color};
-            border-radius: 14px;
-            outline: 1px solid #484848;
-        }}
-        #terminal:after {{
-            position: absolute;
-            width: 100%;
-            height: 100%;
-            content: '';
-            border-radius: 14px;
-            background: rgb(71,77,102);
-            background: linear-gradient(90deg, #804D69 0%, #4E4B89 100%);
-            transform: rotate(-4.5deg);
-            z-index: -1;
-        }}
-        #terminal-header {{
-            position: relative;
-            width: 100%;
-            background-color: #2e2e2e;
-            margin-bottom: 12px;
-            font-weight: bold;
-            border-radius: 14px 14px 0 0;
-            color: {theme_foreground_color};
-            font-size: 18px;
-            box-shadow: inset 0px -1px 0px 0px #4e4e4e,
-                        inset 0px -4px 8px 0px #1a1a1a;
-        }}
-        #terminal-title-tab {{
-            display: inline-block;
-            margin-top: 14px;
-            margin-left: 124px;
-            font-family: sans-serif;
-            padding: 14px 28px;
-            border-radius: 6px 6px 0 0;
-            background-color: {theme_background_color};
-            box-shadow: inset 0px 1px 0px 0px #4e4e4e,
-                        0px -4px 4px 0px #1e1e1e,
-                        inset 1px 0px 0px 0px #4e4e4e,
-                        inset -1px 0px 0px 0px #4e4e4e;
-        }}
-        #terminal-traffic-lights {{
-            position: absolute;
-            top: 24px;
-            left: 20px;
-        }}
-        #terminal-body {{
-            line-height: {line_height}px;
-            padding: 14px;
-        }}
-        {stylesheet}
-    </style>
-    <foreignObject x="0" y="0" width="100%" height="100%">
-        <body xmlns="http://www.w3.org/1999/xhtml">
-            <div id="wrapper">
-                <div id="terminal">
-                    <div id='terminal-header'>
-                        <svg id="terminal-traffic-lights" width="90" height="21" viewBox="0 0 90 21" xmlns="http://www.w3.org/2000/svg">
-                            <circle cx="14" cy="8" r="8" fill="#ff6159"/>
-                            <circle cx="38" cy="8" r="8" fill="#ffbd2e"/>
-                            <circle cx="62" cy="8" r="8" fill="#28c941"/>
-                        </svg>
-                        <div id="terminal-title-tab">{title}</div>
-                    </div>
-                    <div id='terminal-body'>
-                        {code}
-                    </div>
-                </div>
-            </div>
-        </body>
-    </foreignObject>
-</svg>
-"""
 
 _TERM_COLORS = {"256color": ColorSystem.EIGHT_BIT, "16color": ColorSystem.STANDARD}
 
@@ -1045,6 +911,13 @@ class Console:
         """
         if self._force_terminal is not None:
             return self._force_terminal
+
+        if hasattr(sys.stdin, "__module__") and sys.stdin.__module__.startswith(
+            "idlelib"
+        ):
+            # Return False for Idle which claims to be a tty but can't handle ansi codes
+            return False
+
         isatty: Optional[Callable[[], bool]] = getattr(self.file, "isatty", None)
         try:
             return False if isatty is None else isatty()
@@ -1099,16 +972,16 @@ class Console:
         if WINDOWS:  # pragma: no cover
             try:
                 width, height = os.get_terminal_size()
-            except OSError:  # Probably not a terminal
+            except (AttributeError, ValueError, OSError):  # Probably not a terminal
                 pass
         else:
-            try:
-                width, height = os.get_terminal_size(sys.__stdin__.fileno())
-            except (AttributeError, ValueError, OSError):
+            for file_descriptor in _STD_STREAMS:
                 try:
-                    width, height = os.get_terminal_size(sys.__stdout__.fileno())
+                    width, height = os.get_terminal_size(file_descriptor)
                 except (AttributeError, ValueError, OSError):
                     pass
+                else:
+                    break
 
         columns = self._environ.get("COLUMNS")
         if columns is not None and columns.isdigit():
@@ -1311,6 +1184,38 @@ class Console:
         """
         return self._is_alt_screen
 
+    def set_window_title(self, title: str) -> bool:
+        """Set the title of the console terminal window.
+
+        Warning: There is no means within Rich of "resetting" the window title to its
+        previous value, meaning the title you set will persist even after your application
+        exits.
+
+        ``fish`` shell resets the window title before and after each command by default,
+        negating this issue. Windows Terminal and command prompt will also reset the title for you.
+        Most other shells and terminals, however, do not do this.
+
+        Some terminals may require configuration changes before you can set the title.
+        Some terminals may not support setting the title at all.
+
+        Other software (including the terminal itself, the shell, custom prompts, plugins, etc.)
+        may also set the terminal window title. This could result in whatever value you write
+        using this method being overwritten.
+
+        Args:
+            title (str): The new title of the terminal window.
+
+        Returns:
+            bool: True if the control code to change the terminal title was
+                written, otherwise False. Note that a return value of True
+                does not guarantee that the window title has actually changed,
+                since the feature may be unsupported/disabled in some terminals.
+        """
+        if self.is_terminal:
+            self.control(Control.title(title))
+            return True
+        return False
+
     def screen(
         self, hide_cursor: bool = True, style: Optional[StyleType] = None
     ) -> "ScreenContext":
@@ -1422,6 +1327,11 @@ class Console:
             _rendered = self.render(renderable, render_options)
             if style:
                 _rendered = Segment.apply_style(_rendered, style)
+
+            render_height = render_options.height
+            if render_height is not None:
+                render_height = max(0, render_height)
+
             lines = list(
                 islice(
                     Segment.split_and_crop_lines(
@@ -1431,7 +1341,7 @@ class Console:
                         pad=pad,
                     ),
                     None,
-                    render_options.height,
+                    render_height,
                 )
             )
             if render_options.height is not None:
@@ -1901,7 +1811,7 @@ class Console:
         """Prints a rich render of the last exception and traceback.
 
         Args:
-            width (Optional[int], optional): Number of characters used to render code. Defaults to 88.
+            width (Optional[int], optional): Number of characters used to render code. Defaults to 100.
             extra_lines (int, optional): Additional lines of code to render. Defaults to 3.
             theme (str, optional): Override pygments theme used in traceback
             word_wrap (bool, optional): Enable word wrapping of long lines. Defaults to False.
@@ -1947,7 +1857,7 @@ class Console:
         frame = currentframe()
         if frame is not None:
             # Use the faster currentframe where implemented
-            while offset and frame:
+            while offset and frame is not None:
                 frame = frame.f_back
                 offset -= 1
             assert frame is not None
@@ -2066,7 +1976,7 @@ class Console:
                         if self.legacy_windows:
                             try:
                                 use_legacy_windows_render = (
-                                    self.file.fileno() in _STD_STREAMS
+                                    self.file.fileno() in _STD_STREAMS_OUTPUT
                                 )
                             except (ValueError, io.UnsupportedOperation):
                                 pass
@@ -2318,135 +2228,231 @@ class Console:
         clear: bool = True,
         code_format: str = CONSOLE_SVG_FORMAT,
     ) -> str:
-        """Generate an SVG string from the console contents (requires record=True in Console constructor)
+        """
+        Generate an SVG from the console contents (requires record=True in Console constructor).
 
         Args:
+            path (str): The path to write the SVG to.
             title (str): The title of the tab in the output image
             theme (TerminalTheme, optional): The ``TerminalTheme`` object to use to style the terminal
             clear (bool, optional): Clear record buffer after exporting. Defaults to ``True``
             code_format (str): Format string used to generate the SVG. Rich will inject a number of variables
                 into the string in order to form the final SVG output. The default template used and the variables
                 injected by Rich can be found by inspecting the ``console.CONSOLE_SVG_FORMAT`` variable.
-
-        Returns:
-            str: The string representation of the SVG. That is, the ``code_format`` template with content injected.
         """
-        assert (
-            self.record
-        ), "To export console contents set record=True in the constructor or instance"
+
+        from rich.cells import cell_len
+
+        style_cache: Dict[Style, str] = {}
+
+        def get_svg_style(style: Style) -> str:
+            """Convert a Style to CSS rules for SVG."""
+            if style in style_cache:
+                return style_cache[style]
+            css_rules = []
+            color = (
+                _theme.foreground_color
+                if (style.color is None or style.color.is_default)
+                else style.color.get_truecolor(_theme)
+            )
+            bgcolor = (
+                _theme.background_color
+                if (style.bgcolor is None or style.bgcolor.is_default)
+                else style.bgcolor.get_truecolor(_theme)
+            )
+            if style.reverse:
+                color, bgcolor = bgcolor, color
+            if style.dim:
+                color = blend_rgb(color, bgcolor, 0.4)
+            css_rules.append(f"fill: {color.hex}")
+            if style.bold:
+                css_rules.append("font-weight: bold")
+            if style.italic:
+                css_rules.append("font-style: italic;")
+            if style.underline:
+                css_rules.append("text-decoration: underline;")
+            if style.strike:
+                css_rules.append("text-decoration: line-through;")
+
+            css = ";".join(css_rules)
+            style_cache[style] = css
+            return css
 
         _theme = theme or SVG_EXPORT_THEME
 
-        with self._record_buffer_lock:
-            segments = Segment.simplify(self._record_buffer)
-            segments = Segment.filter_control(segments)
-            parts = [(text, style or Style.null()) for text, style, _ in segments]
-            terminal_text = Text.assemble(*parts)
-            lines = terminal_text.wrap(self, width=self.width, overflow="fold")
-            segments = self.render(lines, options=self.options)
-            segment_lines = list(
-                Segment.split_and_crop_lines(
-                    segments, length=self.width, include_new_lines=False
-                )
+        width = self.width
+        char_height = 20
+        char_width = char_height * 0.61
+        line_height = char_height * 1.22
+
+        margin_top = 1
+        margin_right = 1
+        margin_bottom = 1
+        margin_left = 1
+
+        padding_top = 40
+        padding_right = 8
+        padding_bottom = 12
+        padding_left = 8
+
+        padding_width = padding_left + padding_right
+        padding_height = padding_top + padding_bottom
+        margin_width = margin_left + margin_right
+        margin_height = margin_top + margin_bottom
+
+        text_backgrounds: List[str] = []
+        text_group: List[str] = []
+        classes: Dict[str, int] = {}
+        style_no = 1
+
+        def escape_text(text: str) -> str:
+            """HTML escape text and replace spaces with nbsp."""
+            return escape(text).replace(" ", "&#160;")
+
+        def make_tag(
+            name: str, content: Optional[str] = None, **attribs: object
+        ) -> str:
+            """Make a tag from name, content, and attributes."""
+
+            def stringify(value: object) -> str:
+                if isinstance(value, (float)):
+                    return format(value, "g")
+                return str(value)
+
+            tag_attribs = " ".join(
+                f'{k.lstrip("_").replace("_", "-")}="{stringify(v)}"'
+                for k, v in attribs.items()
+            )
+            return (
+                f"<{name} {tag_attribs}>{content}</{name}>"
+                if content
+                else f"<{name} {tag_attribs}/>"
             )
 
-            fragments: List[str] = []
-            theme_foreground_color = _theme.foreground_color.hex
-            theme_background_color = _theme.background_color.hex
-
-            theme_foreground_css = f"color: {theme_foreground_color}; text-decoration-color: {theme_foreground_color};"
-            theme_background_css = f"background-color: {theme_background_color};"
-
-            theme_css = theme_foreground_css + theme_background_css
-
-            styles: Dict[str, int] = {}
-            styles[theme_css] = 1
-
-            for line in segment_lines:
-                line_spans = []
-                for segment in line:
-                    text, style, _ = segment
-                    text = escape(text)
-                    if style:
-                        rules = style.get_html_style(_theme)
-                        if style.link:
-                            text = f'<a href="{style.link}">{text}</a>'
-
-                        if style.blink or style.blink2:
-                            text = f'<span class="blink">{text}</span>'
-
-                        # If the style doesn't contain a color, we still
-                        # need to make sure we output the default foreground color
-                        # from the TerminalTheme.
-                        if not style.reverse:
-                            foreground_css = theme_foreground_css
-                            background_css = theme_background_css
-                        else:
-                            foreground_css = f"color: {theme_background_color}; text-decoration-color: {theme_background_color};"
-                            background_css = (
-                                f"background-color: {theme_foreground_color};"
-                            )
-
-                        if style.color is None:
-                            rules += f";{foreground_css}"
-                        if style.bgcolor is None:
-                            rules += f";{background_css}"
-
-                        style_number = styles.setdefault(rules, len(styles) + 1)
-                        text = f'<span class="r{style_number}">{text}</span>'
-                    else:
-                        text = f'<span class="r1">{text}</span>'
-                    line_spans.append(text)
-
-                fragments.append(f"<div>{''.join(line_spans)}</div>")
-
-            stylesheet_rules = []
-            for style_rule, style_number in styles.items():
-                if style_rule:
-                    stylesheet_rules.append(f".r{style_number} {{{ style_rule }}}")
-            stylesheet = "\n".join(stylesheet_rules)
-
+        with self._record_buffer_lock:
+            segments = list(
+                Segment.filter_control(
+                    Segment.simplify(self._record_buffer),
+                )
+            )
             if clear:
                 self._record_buffer.clear()
 
-        # These values are the ones that I found to work well after experimentation.
-        # Many of them can be tweaked, but too much variation from these values could
-        # result in visually broken output/clipping issues.
-        terminal_padding = 12
-        font_size = 18
-        line_height = font_size + 4
-        code_start_y = 60
-        required_code_height = line_height * len(lines)
-        margin = 140
-
-        # Monospace fonts are generally around 0.5-0.55 width/height ratio, but I've
-        # added extra width to ensure that the output SVG is big enough.
-        monospace_font_width_scale = 0.60
-
-        # This works out as a good heuristic for the final size of the drawn terminal.
-        terminal_height = required_code_height + code_start_y
-        terminal_width = (
-            self.width * monospace_font_width_scale * font_size
-            + 2 * terminal_padding
-            + self.width
+        unique_id = "terminal-" + str(
+            zlib.adler32(
+                ("".join(segment.text for segment in segments)).encode(
+                    "utf-8", "ignore"
+                )
+                + title.encode("utf-8", "ignore")
+            )
         )
-        total_height = terminal_height + 2 * margin
-        total_width = terminal_width + 2 * margin
+        y = 0
+        for y, line in enumerate(Segment.split_and_crop_lines(segments, length=width)):
+            x = 0
+            for text, style, _control in line:
+                style = style or Style()
+                rules = get_svg_style(style)
+                if rules not in classes:
+                    classes[rules] = style_no
+                    style_no += 1
+                class_name = f"r{classes[rules]}"
 
-        rendered_code = code_format.format(
-            code="\n".join(fragments),
-            total_height=total_height,
-            total_width=total_width,
-            theme_foreground_color=theme_foreground_color,
-            theme_background_color=theme_background_color,
-            margin=margin,
-            font_size=font_size,
+                if style.reverse:
+                    has_background = True
+                    background = (
+                        _theme.foreground_color.hex
+                        if style.color is None
+                        else style.color.get_truecolor(_theme).hex
+                    )
+                else:
+                    bgcolor = style.bgcolor
+                    has_background = bgcolor is not None and not bgcolor.is_default
+                    background = (
+                        _theme.background_color.hex
+                        if style.bgcolor is None
+                        else style.bgcolor.get_truecolor(_theme).hex
+                    )
+
+                text_length = cell_len(text)
+                if has_background:
+                    text_backgrounds.append(
+                        make_tag(
+                            "rect",
+                            fill=background,
+                            x=x * char_width,
+                            y=y * line_height,
+                            width=char_width * text_length + 1,
+                            height=line_height + 1,
+                        )
+                    )
+
+                if text != " " * len(text):
+                    text_group.append(
+                        make_tag(
+                            "text",
+                            escape_text(text),
+                            _class=f"{unique_id}-{class_name}",
+                            x=x * char_width,
+                            y=y * line_height + char_height,
+                            textLength=char_width * len(text),
+                        )
+                    )
+                x += cell_len(text)
+
+        styles = "\n".join(
+            f".{unique_id}-r{rule_no} {{ {css} }}" for css, rule_no in classes.items()
+        )
+        backgrounds = "".join(text_backgrounds)
+        matrix = "".join(text_group)
+
+        terminal_width = ceil(width * char_width + padding_width)
+        terminal_height = (y + 1) * line_height + padding_height
+        chrome = make_tag(
+            "rect",
+            fill=_theme.background_color.hex,
+            stroke="rgba(255,255,255,0.35)",
+            stroke_width="1",
+            x=margin_left,
+            y=margin_top,
+            width=terminal_width,
+            height=terminal_height,
+            rx=12,
+        )
+
+        title_color = _theme.foreground_color.hex
+        if title:
+            chrome += make_tag(
+                "text",
+                escape_text(title),
+                _class=f"{unique_id}-title",
+                fill=title_color,
+                text_anchor="middle",
+                x=terminal_width // 2,
+                y=margin_top + char_height + 6,
+            )
+        chrome += f"""
+            <g transform="translate(30,24)">
+            <circle cx="0" cy="0" r="7" fill="#ff5f57"/>
+            <circle cx="22" cy="0" r="7" fill="#febc2e"/>
+            <circle cx="44" cy="0" r="7" fill="#28c840"/>
+            </g>
+        """
+
+        svg = code_format.format(
+            unique_id=unique_id,
+            char_width=char_width,
+            char_height=char_height,
             line_height=line_height,
-            title=title,
-            stylesheet=stylesheet,
+            width=terminal_width + margin_width,
+            height=terminal_height + margin_height,
+            terminal_x=margin_left + padding_left,
+            terminal_y=margin_top + padding_top,
+            styles=styles,
+            chrome=chrome,
+            backgrounds=backgrounds,
+            matrix=matrix,
         )
-
-        return rendered_code
+        return svg
 
     def save_svg(
         self,
@@ -2469,12 +2475,27 @@ class Console:
                 injected by Rich can be found by inspecting the ``console.CONSOLE_SVG_FORMAT`` variable.
         """
         svg = self.export_svg(
-            title=title, theme=theme, clear=clear, code_format=code_format
+            title=title,
+            theme=theme,
+            clear=clear,
+            code_format=code_format,
         )
         with open(path, "wt", encoding="utf-8") as write_file:
             write_file.write(svg)
 
 
+def _svg_hash(svg_main_code: str) -> str:
+    """Returns a unique hash for the given SVG main code.
+
+    Args:
+        svg_main_code (str): The content we're going to inject in the SVG envelope.
+
+    Returns:
+        str: a hash of the given content
+    """
+    return str(zlib.adler32(svg_main_code.encode()))
+
+
 if __name__ == "__main__":  # pragma: no cover
     console = Console(record=True)
 
